"""
python v3.7.9
@Project: hotpot
@File   : bundle.py
@Author : Zhiyuan Zhang
@Date   : 2023/3/22
@Time   : 3:18
"""
import copy
from os import PathLike
from typing import *
from pathlib import Path
import numpy as np
from tqdm import tqdm
from openbabel import pybel as pb
import hotpot.cheminfo as ci
from hotpot.tools import check_path
import multiprocessing as mp


# Typing AnnotationsG
GraphFormatName = Literal['Pytorch', 'numpy']

feature_formats = {
    'basic': ['atomic_number', 's', 'p', 'f']
}


class MolBundle:
    """"""

    def __init__(self, mols: Union[list[ci.Molecule], Generator[ci.Molecule, None, None]] = None):
        self.mols = mols
        self._data = {}

    def __repr__(self):
        return f"MolBundle[{', '.join([str(m) for m in self.mols])}]"

    def __iter__(self):
        return iter(tqdm(self.mols))

    def __add__(self, other: Union['MolBundle', ci.Molecule]):
        """"""
        if isinstance(other, MolBundle):
            return MolBundle(self.to_list() + other.to_list())
        elif isinstance(other, ci.Molecule):
            return MolBundle(self.to_list() + [other])
        else:
            raise TypeError('the MolBundle is only allowed to add with Molecule or MolBundle object')

    def __len__(self):
        return len(self.mols)

    def __getitem__(self, item: int):
        return self.mols[item]

    def __bundle_mol_indices_with_same_attr(self, attr_name: str):
        self.to_list()

        dict_attrs = {}
        for i, mol in enumerate(self):
            list_indices = dict_attrs.setdefault(getattr(mol, attr_name), [])
            list_indices.append(i)

        if not dict_attrs:
            return 0
        if len(dict_attrs) == 1:
            return getattr(self.mols[0], attr_name)
        else:
            return dict_attrs

    @property
    def atom_num(self) -> Dict[int, List[int]]:
        """
        Notes:
            if the Bundle is a generator, convert to a list of Molecule first.
        Returns:
            returns a dict with the key is the number of the atoms and the key is the indices of Molecules
        """
        return self.__bundle_mol_indices_with_same_attr('atom_num')

    @property
    def atomic_numbers(self):
        return self.__bundle_mol_indices_with_same_attr('atomic_numbers')

    @classmethod
    def read_from_dir(
            cls, fmt: str,
            read_dir: Union[str, PathLike],
            match_pattern: str = '*',
            generate: bool = False,
            ranges: Union[Sequence[int], range] = None,
            condition: Callable = None,
            num_proc: int = None
    ):
        """
        Read Molecule objects from a directory.

        Args:
            fmt(str): read file with the specified-format method.
            read_dir(str|PathLike): the directory all file put
            match_pattern(str): the file name pattern
            generate(bool): read file to a generator of Molecule object
            ranges(Sequence[int]|range): A list or range of integers representing the indices of
                the input files to read. Defaults to None.
            condition(Callable): A callable object that takes two arguments (the path of the input file
                and the corresponding Molecule object) and returns a boolean value indicating whether to include the
                Molecule object in the output. Defaults to None.
            num_proc: the number of process to read

        Returns:
            List(Molecule) or Generator(Molecule)
        """
        def read_mol(pm: Path, conn):

            try:
                mol = ci.Molecule.read_from(pm, fmt)

                # the OBMol, OBAtom, OBBond are C++ objects wrapped by SWIG,
                # they can't be pickle, so pop them from Molecule and convert their info to .mol2 string
                pmol = pb.Molecule(mol.ob_mol_pop())
                script = pmol.write('mol2')  # Write to the mol2 script to allow it to pickle

                # communicate with the main process
                conn.send((mol, script, pm))

            except AttributeError:
                conn.send((None, None, pm))

        def mol_generator():
            nonlocal read_dir
            for i, path_mol in enumerate(read_dir.glob(match_pattern)):

                if not ranges or i in ranges:
                    mol = ci.Molecule.read_from(path_mol, fmt)
                else:
                    continue

                if mol and (not condition or condition(path_mol, mol)):
                    yield mol

        def mol_mp_generator():
            mols_info = []

            parent_conn, child_conn = mp.Pipe()
            ps = []  # list of Process: Queue pairs

            for i, path_file in enumerate(read_dir.glob(match_pattern)):

                # if the number of process more than num_proc, get the read molecule info and stop to start new process
                while len(ps) >= num_proc:
                    for p in ps:
                        if not p.is_alive():
                            mols_info.append(parent_conn.recv())
                            p.terminate()
                            ps.remove(p)

                # When received some Molecule info, reorganize these info and yield
                while mols_info:
                    mol, script, pf = mols_info.pop()  # hotpot Molecule, mol2_script, path_file of Molecule

                    # if get a valid Molecule info, re-wrap to be hotpot Molecule
                    if mol and script:
                        pmol = pb.readstring('mol2', script)  # pybel Molecule object
                        mol.ob_mol_rewrap(pmol.OBMol)  # re-wrap OBMol by hotpot Molecule

                        # if the reorganized Molecule is expected, yield
                        if not condition or condition(mol, pf):
                            yield mol

                # Start new process to read Molecule from file
                if not ranges or i in ranges:
                    p = mp.Process(target=read_mol, args=(path_file, child_conn))
                    p.start()
                    ps.append(p)

            # After all path_file have pass into process to read
            while ps:
                for p in ps:
                    if not p.is_alive():
                        mols_info.append(parent_conn.recv())
                        p.terminate()
                        ps.remove(p)

            for mol, script, pf in mols_info:
                # if get a valid Molecule info, re-wrap to be hotpot Molecule
                if mol and script:
                    pmol = pb.readstring('mol2', script)  # pybel Molecule object
                    mol.ob_mol_rewrap(pmol.OBMol)  # re-wrap OBMol by hotpot Molecule

                    # if the reorganized Molecule is expected, yield
                    if not condition or condition(mol, pf):
                        yield mol

        if isinstance(read_dir, str):
            read_dir = Path(read_dir)
        elif not isinstance(read_dir, PathLike):
            raise TypeError(f'the read_dir should be a str or PathLike, instead of {type(read_dir)}')

        generator = mol_mp_generator() if num_proc else mol_generator()

        if generate:
            return cls(generator)
        else:
            return cls([m for m in tqdm(generator, 'reading molecules')])

    def graph_representation(self, *feature_names) -> Generator[Union[str, np.ndarray, np.ndarray], None, None]:
        """ Transform molecules to their graph representation """
        for mol in self.mols:
            yield mol.graph_representation(*feature_names)

    def gaussian(
            self, g16root: Union[str, PathLike], dir_out: Union[str, PathLike],
            link0: Union[str, List[str]], route: Union[str, List[str]],
            dir_err: Optional[Union[str, PathLike]] = None,
            dir_chk: Optional[Union[str, PathLike]] = None,
            clean_configure: bool = True,
            perturb_kwargs: Optional[Union[Dict[str, Any], List[Dict[str, Any]]]] = None,
            *args, **kwargs
    ) -> None:
        """
        Run Gaussian16 calculations on the Molecule objects.

        Args:
            g16root (Union[str, PathLike]): The path to the Gaussian16 root directory.
            dir_out (Union[str, PathLike]): The path to the directory to output the log files.
            link0 (Union[str, List[str]]): The link0 information for Gaussian16 calculations.
            route (Union[str, List[str]]): The route information for Gaussian16 calculations.
            dir_err (Optional[Union[str, PathLike]], optional): The path to the directory to output the error files.
                Defaults to None.
            dir_chk (Optional[Union[str, PathLike]], optional): The path to the directory to store the .chk files.
                Defaults to None.
            clean_configure (bool, optional): A flag indicating whether to clean the configuration before perturbing
                the molecule or lattice. Defaults to True.
            perturb_kwargs (Optional[Union[Dict[str, Any], List[Dict[str, Any]]]], optional): The parameters for
                perturbing the molecule or lattice. Defaults to None.
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.

        Returns:
            None
        """
        # Check and process paths
        g16root: Path = check_path(g16root, file_or_dir='dir')
        dir_out: Path = check_path(dir_out, mkdir=True)
        dir_err: Optional[Path] = check_path(dir_err, mkdir=True)
        dir_chk: Optional[Path] = check_path(dir_chk, mkdir=True)

        for mol in self.mols:
            assert isinstance(mol, ci.Molecule)

            # Clean before perturb configures
            if clean_configure:
                mol.clean_configures()

            # Assign the dirs
            # assign the dir to put out.log files for the mol
            dir_out_mol = dir_out.joinpath(mol.identifier)
            if not dir_out_mol.exists():
                dir_out_mol.mkdir()

            # assign the dir to put err.log files for the mol
            if dir_err:
                dir_err_mol = dir_err.joinpath(mol.identifier)
                if not dir_err_mol.exists():
                    dir_err_mol.mkdir()
            else:
                dir_err_mol = None

            # Performing the molecule or lattice perturb
            if isinstance(perturb_kwargs, Dict):
                perturb_kwargs = [perturb_kwargs]  # Wrap it into a list

            if isinstance(perturb_kwargs, List) and all(isinstance(pk, dict) for pk in perturb_kwargs):
                for pk in perturb_kwargs:
                    pk['inplace'] = True  # Force to inplace
                    mol.perturb_mol_lattice(**pk)
            elif perturb_kwargs is not None:
                ValueError('The perturb_kwargs should be a dict or list of dict')

            # Running the gaussian16
            for config_idx in range(mol.configure_number):
                mol.configure_select(config_idx)

                # Reorganize the arguments for each configures
                path_out = dir_out_mol.joinpath(f'{config_idx}.log')
                path_err = dir_err_mol.joinpath(f'{config_idx}.err') if dir_err else None

                if dir_chk:  # for dir_chk
                    dir_chk_mol = dir_chk.joinpath(mol.identifier)
                    dir_chk_mol.mkdir(exist_ok=True)

                    if isinstance(link0, str):
                        link0_cfg = [f'chk={str(dir_chk_mol)}/{config_idx}.chk', link0]
                    elif isinstance(link0, list):
                        link0_cfg = copy.copy(link0)
                        link0_cfg.insert(0, f'chk={str(dir_chk_mol)}/{config_idx}.chk')
                    else:
                        TypeError('the link0 should be str or list of str!')
                else:
                    link0_cfg = link0

                mol.gaussian(
                    g16root=g16root,
                    link0=link0_cfg,
                    route=route,
                    path_log_file=str(path_out.absolute()),
                    path_err_file=str(path_err) if path_err else None,
                    *args, **kwargs
                )

    def merge_conformers(self):
        """
        Get the sum of conformers for all molecule in the mol bundle "self.mols"
        This method can only be successfully executed
        when all molecules in the molecular bundle can be added to each other
        Returns:
            a Molecule object with all of conformers in the self.mols
        """
        atomic_numbers = self.atomic_numbers

        if isinstance(atomic_numbers, tuple):
            return sum(self.mols[1:], start=self.mols[0])
        elif isinstance(atomic_numbers, dict):
            mol_array = np.array(self.mols)
            return MolBundle([mol_array[i].sum() for ans, i in self.atomic_numbers.items()])

    def merge_atoms_same_mols(self):
        """ Merge Molecules with same atoms to a MixSameAtomMol """
        bundle = self.to_mix_mols()
        atom_num = bundle.atom_num

        if isinstance(atom_num, tuple):
            return sum(bundle.mols[1:], start=bundle.mols[0])
        elif isinstance(atom_num, dict):
            mol_array = np.array(bundle.mols)
            return MolBundle([mol_array[i].sum() for ans, i in atom_num.items()])

    def to_dpmd_sys(self, sys_root):
        """"""

    def to_list(self) -> List[ci.Molecule]:
        """ Convert the molecule container (self.mol) to list """
        if isinstance(self.mols, Generator):
            self.mols = list(self)

        return self.mols

    def to_mix_mols(self):
        """
        Return a new MolBundle, in which Molecule objects in container are converted to MixSameAtomMol
        Returns:
            MolBundle(MixSameAtomMol)
        """
        return MolBundle([m.to_mix_mol() if not isinstance(m, ci.MixSameAtomMol) else m for m in self])

    def to_mols(self):
        """
        Return a new MolBundle, in which MixSameAtomMol objects in container are converted to Molecule
        Returns:
            MolBundle(Molecule)
        """
        return MolBundle([m.to_mol() if isinstance(m, ci.MixSameAtomMol) else m for m in self])



"""
python v3.7.9
@Project: hotpot
@File   : quantum.py
@Author : Zhiyuan Zhang
@Date   : 2023/3/20
@Time   : 2:44
"""
import os
import resource
import subprocess
import io
import cclib
from typing import *


class Gaussian:
    """
    A class for setting up and running Gaussian 16 calculations.

    Attributes:
        g16root (str): The path to the Gaussian 16 root directory.

    """
    def __init__(self, g16root: Union[str, os.PathLike], report_set_resource_error: bool = False):
        """
        This method sets up the required environment variables and resource limits for Gaussian 16.
        Args:
            g16root (Union[str, os.PathLike]): The path to the Gaussian 16 root directory.

        Raises:
            TypeError: If `g16root` is not a string or a path-like object.
        """
        if isinstance(g16root, str):
            self.g16root = g16root
        elif isinstance(g16root, os.PathLike):
            self.g16root = str(g16root)
        else:
            raise TypeError('the g16root should be str or os.PathLike type!')

        self.envs = self._set_environs()
        self._set_resource_limits(report_set_resource_error)

        self.data = None

    def _set_environs(self):
        """Sets up the environment variables required for running Gaussian 16.

        This method sets the environment variables required for Gaussian 16 to function correctly. If the
        `g16root` attribute is not set, the method sets it to the user's home directory.

        Returns:
            Dict[str, str]: A dictionary of the updated environment variables."""

        if self.g16root:
            g16root = self.g16root
        else:
            g16root = os.path.expanduser("~")

        # Setting environment for gaussian 16
        gr = g16root
        env_vars = {
            'g16root': gr,
            'GAUSS_EXEDIR': f"{gr}/g16/bsd:{gr}/g16",
            'GAUSS_LEXEDIR': f"{gr}/g16/linda-exe",
            'GAUSS_ARCHDIR': f"{gr}/g16/arch",
            'GAUSS_BSDDIR': f"{gr}/g16/bsd",
            'GV_DIR': f"{gr}/gv",
            'PATH': f"{os.environ['PATH']}:{gr}/gauopen:{gr}/g16/bsd:{gr}/g16",
            'PERLLIB': f"{os.environ['PERLLIB']}:{gr}/gauopen:{gr}/g16/bsd:{gr}/g16" if 'PERLLIB' in os.environ else f"{gr}/gauopen:{gr}/g16/bsd:{gr}/g16",
            'PYTHONPATH': f"{os.environ['PYTHONPATH']}:{gr}/gauopen:{gr}/g16/bsd:{gr}/g16" if 'PYTHONPATH' in os.environ else f"{gr}/gauopen:{gr}/g16/bsd:{gr}/g16",
            '_DSM_BARRIER': 'SHM',
            'LD_LIBRARY64_PATH': f"{gr}/g16/bsd:{gr}/gv/lib:{os.environ['LD_LIBRARY64_PATH']}" if 'LD_LIBRARY64_PATH' in os.environ else "",
            'LD_LIBRARY_PATH': f"{gr}/g16/bsd:{os.environ['LD_LIBRARY_PATH']}:{gr}/gv/lib" if 'LD_LIBRARY_PATH' in os.environ else f"{gr}/g16/bsd:{gr}/gv/lib",
            'G16BASIS': f"{gr}/g16/basis",
            'PGI_TERM': 'trace,abort'
        }

        # Merge the environment variables with the current environment
        updated_env = os.environ.copy()
        updated_env.update(env_vars)

        return updated_env

    @staticmethod
    def _set_resource_limits(report_error: bool):
        """Sets resource limits for the Gaussian 16 process to avoid system crashes.

        This method sets resource limits for the Gaussian 16 process to avoid system crashes. Specifically,
        it sets the limits for the following resources: core dump size, data segment size, file size,
        locked-in-memory address space, resident set size, number of open files, stack size, CPU time,
        and number of processes.
        """
        try:
            resource.setrlimit(resource.RLIMIT_CORE, (0, 0))
        except ValueError:
            if report_error:
                print(RuntimeWarning('Unable to raise the RLIMIT_CORE limit.'))

        try:
            resource.setrlimit(resource.RLIMIT_DATA, (resource.RLIM_INFINITY, resource.RLIM_INFINITY))
        except ValueError:
            if report_error:
                print(RuntimeWarning('Unable to raise the RLIMIT_DATA limit.'))

        try:
            resource.setrlimit(resource.RLIMIT_FSIZE, (resource.RLIM_INFINITY, resource.RLIM_INFINITY))
        except ValueError:
            if report_error:
                print(RuntimeWarning('Unable to raise the RLIMIT_FSIZE limit.'))

        try:
            resource.setrlimit(resource.RLIMIT_MEMLOCK, (resource.RLIM_INFINITY, resource.RLIM_INFINITY))
        except ValueError:
            if report_error:
                print(RuntimeWarning('Unable to raise the RLIMIT_MEMLOCK limit.'))

        try:
            resource.setrlimit(resource.RLIMIT_RSS, (resource.RLIM_INFINITY, resource.RLIM_INFINITY))
        except ValueError:
            if report_error:
                print(RuntimeWarning('Unable to raise the RLIMIT_RSS limit.'))

        try:
            resource.setrlimit(resource.RLIMIT_NOFILE, (resource.RLIM_INFINITY, resource.RLIM_INFINITY))
        except ValueError:
            if report_error:
                print(RuntimeWarning('Unable to raise the RLIMIT_NOFILE limit.'))

        try:
            resource.setrlimit(resource.RLIMIT_STACK, (resource.RLIM_INFINITY, resource.RLIM_INFINITY))
        except ValueError:
            if report_error:
                print(RuntimeWarning('Unable to raise the RLIMIT_STACK limit.'))

        try:
            resource.setrlimit(resource.RLIMIT_CPU, (resource.RLIM_INFINITY, resource.RLIM_INFINITY))
        except ValueError:
            if report_error:
                print(RuntimeWarning('Unable to raise the RLIMIT_CPU limit.'))

        try:
            resource.setrlimit(resource.RLIMIT_NPROC, (resource.RLIM_INFINITY, resource.RLIM_INFINITY))
        except ValueError:
            if report_error:
                print(RuntimeWarning('Unable to raise the RLIMIT_NPROC limit.'))

    def run(self, script: str, *args, **kwargs):
        """Runs the Gaussian 16 process with the given script and additional arguments.

        This method sets up the required environment variables and resource limits for Gaussian 16 before
        running the process using `subprocess.Popen`. It takes an input script and any additional arguments
        to pass to `Popen`, and returns a tuple of the standard output and standard error of the process.

        Args:
            script (str): The input script for the Gaussian 16 process.
            *args: Additional arguments to pass to subprocess.Popen.
            **kwargs: Additional keyword arguments to pass to subprocess.Popen.

        Returns:
            Tuple[str, str]: A tuple of the standard output and standard error of the process.
        """
        # Run Gaussian using subprocess
        g16_process = subprocess.Popen(
            'g16', bufsize=-1, stdin=subprocess.PIPE,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE,
            env=self.envs, universal_newlines=True
        )
        stdout, stderr = g16_process.communicate(script)

        self.parse_log(stdout)

        return stdout, stderr

    def parse_log(self, stdout: str):
        """ Parse the gaussian log file and save them into self """
        string_buffer = io.StringIO(stdout)
        self.data: cclib.parser.data.ccData_optdone_bool = cclib.ccopen(string_buffer).parse()

    @property
    def molecule_setter_dict(self):
        """ Prepare the property dict for Molecule setters """
        return {
            'atoms.partial_charge': self.data.atomcharges['mulliken'],
            'energy': self.data.scfenergies,
            'spin': self.data.mult,
            'charge': self.data.charge,
            'mol_orbital_energies': self.data.moenergies  # eV
        }




"""
python v3.7.9
@Project: hotpot
@File   : base.py
@Author : Zhiyuan Zhang
@Date   : 2023/4/24
@Time   : 3:43
"""
import os
import re
from os.path import join as ptj
import json

import numpy as np
from lammps import PyLammps, Atom
from typing import *
import hotpot


class HpLammps:
    """
    A wrapper to run LAMMPS tasks
    """
    def __init__(self, main, **kwargs):
        """

        Args:
            main: the main Molecule object

        Keyword Args:
            work_dir: the work dir, where the io operations are performed
        """
        self._data = {
            'main': main,
            'pylmp': PyLammps()
        }  # store any data
        self._data.update(kwargs)

    def __call__(self, cmd: str):
        self.command(cmd)

    # def __dir__(self) -> Iterable[str]:
    #     return self.pylmp.__dir__() + ["commands_list", "commands_string", 'read_main_data', "script"]
    #
    # def __getattr__(self, item):
    #     return self.pylmp.__getattr__(item)

    def atom(self, idx: int):
        """ retrieve the LAMMPS atom object by atom index """
        return Atom(self, idx)

    @property
    def atoms(self):
        return [self.atom(i) for i in range(self.atom_number)]

    @property
    def atom_number(self):
        return len(self.pylmp.atoms)

    @property
    def box(self):
        return np.array([
            [self.eval('xlo'), self.eval('xhi')],
            [self.eval('ylo'), self.eval('yhi')],
            [self.eval('zlo'), self.eval('zhi')]
        ])

    def close(self):
        self.pylmp.close()
        del self

    def command(self, cmd: str):
        self.pylmp.command(cmd)

    def commands_list(self, list_cmd: List[str]):
        for cmd in list_cmd:
            self.command(cmd)

    def commands_string(self, multicod: str):
        cmd_list = multicod.split('\n')
        for cmd in cmd_list:
            self.command(cmd.strip())

    @property
    def computes(self):
        return self.pylmp.computes

    @property
    def cryst_matrix(self):
        xl = self.eval('xhi')-self.eval('xlo')  # length of the crystal box in the x axis
        yl = self.eval('xhi')-self.eval('xlo')  # length of the crystal box in the y axis
        zl = self.eval('xhi')-self.eval('xlo')  # length of the crystal box in the z axis
        xy = self.eval('xy')  # the project v_b to x axis
        xz = self.eval('xz')  # the project v_c to x axis
        yz = self.eval('yz')  # the project v_c to y axis
        return np.array([[xl, 0., 0.], [xy, yl, 0.], [xz, yz, zl]])

    def data_to_labelmap(self, script: str, offset: int = 0):
        """ Convert the LAMMPS data format script (string) to labelmap dict """
        pattern = re.compile(r"[A-Z].+s")
        data_body_headers = pattern.findall(script)
        masses_idx = data_body_headers.index('Masses')
        masses_body_contents: List[str] = pattern.split(script)[masses_idx+1].split('\n')

        type_map = self._data.setdefault('type_map', {})
        for line in masses_body_contents:
            line = line.strip()
            if line:
                type_num, _, _, type_label = line.split()
                type_num = int(type_num) + offset  # offset
                former_label = type_map.get(type_num)

                if not former_label:  # if the atom type have not been recorded
                    type_map[type_num] = type_label
                else:
                    # Never allow to change the type_map for defined atom type
                    # if the current type label is different from the former, raise error
                    # if the current type label is same with the former, keeping still.
                    if former_label != type_label:
                        raise RuntimeError(
                            f'the the type_map for atom type {type_num} is attempt to change '
                            f'from {former_label} to {type_label}, Never allowed!!'
                        )

        return type_map

    @property
    def dumps(self):
        return self.pylmp.dumps

    def eval(self, expr):
        return self.pylmp.eval(expr)

    @property
    def fixes(self):
        return self.pylmp.fixes

    @property
    def groups(self):
        return self.pylmp.groups

    def file(self, filepath: str):
        self.pylmp.file(filepath)

    @property
    def lmp(self):
        return self.pylmp.lmp

    @property
    def main(self):
        return self._data.get('main')

    @property
    def pylmp(self):
        return self._data.get('pylmp')

    def read_main_data(
            self, add: str = None, offset: Sequence[int] = (), shift: Sequence[int] = (),
            extra_atom_types=0, extra_bond_types=0, extra_angle_types=0, extra_dihedral_types=0,
            extra_improper_types=0, extra_bond_per_atom=0, extra_angle_per_atom=0, extra_dihedral_per_atom=0,
            extra_improper_per_atom=0, extra_special_per_atom=0, group: str = None, nocoeff: bool = False,
            fix: str = None
    ):
        """"""
        path_main_data = os.path.join(self.work_dir, 'main.data')

        # to the main.data file
        self.main.writefile('lmpdat', path_main_data, retrieve_script=False)
        # self.data_to_labelmap(script, atom_offset)

        # read to LAMMPS
        cmd = f'read_data {path_main_data}'

        # add keywords to specify the extra atoms
        if add:
            cmd += f' add {add}'
        if offset:
            assert len(offset) == 5
            cmd += f' offset ' + ' '.join(offset)
        if shift:
            assert len(shift) == 3
            cmd += f' shift ' + ' '.join(shift)
        if extra_atom_types:
            cmd += f' extra/atom/types {extra_atom_types}'
        if extra_bond_types:
            cmd += f' extra/bond/types {extra_bond_types}'
        if extra_angle_types:
            cmd += f' extra/angle/types {extra_angle_types}'
        if extra_dihedral_types:
            cmd += f' extra/dihedral_types {extra_dihedral_types}'
        if extra_improper_types:
            cmd += f' extra/improper/types {extra_improper_types}'
        if extra_bond_per_atom:
            cmd += f' extra/bond/per/atom {extra_bond_per_atom}'
        if extra_angle_per_atom:
            cmd += f' extra/angle/per/atom {extra_angle_per_atom}'
        if extra_dihedral_per_atom:
            cmd += f' extra/dihedral_per_atom {extra_dihedral_per_atom}'
        if extra_improper_per_atom:
            cmd += f' extra/improper/per/atom {extra_improper_per_atom}'
        if extra_special_per_atom:
            cmd += f' extra/special/per/atom {extra_special_per_atom}'
        if group:
            cmd += f' group {group}'
        if nocoeff:
            cmd += f' group true'
        if fix:
            cmd += f' fix {fix}'

        self.command(cmd)

    def run(self, *args, **kwargs):
        self.pylmp.run(*args, **kwargs)

    @property
    def runs(self):
        return self.pylmp.runs

    def script(self):
        path_tmp_script = os.path.join(os.getcwd(), 'script.in')
        self.pylmp.write_script(path_tmp_script)
        with open(path_tmp_script) as file:
            script = file.read()

        return script

    @property
    def type_map(self):
        return self._data.get('type_map')

    @property
    def variables(self):
        return self.pylmp.variables

    @property
    def version(self):
        return self.pylmp.version()

    @property
    def work_dir(self):
        work_dir = self._data.get('work_dir', os.getcwd())
        if not work_dir:
            os.mkdir(work_dir)
        return work_dir

    def write_script(self, filepath: str):
        self.pylmp.write_script(filepath)


class LmpTask:
    """ The base class for performing LAMMPS tasks with Molecule objects """
    def __init__(self, mol: 'ci.Molecule'):
        self.mol = mol


import hotpot.cheminfo as ci



"""
python v3.7.9
@Project: hotpot
@File   : _io.py
@Author : Zhiyuan Zhang
@Date   : 2023/3/14
@Time   : 4:18
"""
import os
import re
from pathlib import Path
from os import PathLike
from typing import *
from abc import ABCMeta, abstractmethod
from copy import copy
import io
from io import IOBase
from openbabel import pybel
import cclib
import numpy as np
import hotpot.cheminfo as ci

"""
Notes:
This module contain Classes to IO between the Molecule object and various file formats.
the main classes are:
    - Reader: from Formatted files to Molecule, called by the Molecule.read() method.
    - Writer: from Molecule to formatted files, called by the Molecule.write() method.
    - Dumper: from Molecule to formatted Literals(str or bytes), called by the Molecule.dump() method.
    - Parser: from formatted Literals to the Molecule obj, called by the Molecule.parse() method.
all of them are inherit from the _IOBase class.

For the implementation of dump and parse:
    - some third-part package are used, like openbabel, cclib and so on.
    - there are some self own IO function be defined too.
    - the user also custom and register to the Dumper and Parser too.

When implementing the IO: 
    1) the IO class will firstly check whether a custom IO function is defined and register, if the IO function 
        are defined and registered, the IO are implemented by the registered.
    2) else, the IO class will try to call, in turn, try to call some third-part packages,
    3) finally, if all third-part packages fail to complement IO, Raise the IOError !!!

Following the steps to customise your IO function:
    1) determine which IO operation(read, write, dump or parse) you want to define， import the relevant IO class
        into your own python modules.
    2) define the IO function which customises your IO implementation, the custom function should meet the base
        requirements demand by the corresponding IO class. When the IO functions are defined, applying the
        `IOClass`.register decorator to register the IO function into the `IOClass`, the `IOClass`.register should
        pass some args. This is a example:
        --------------------------------- Example ------------------------------------
        Examples:
        # importing the relevant IO classes
        from hotpot.io import Dumper, Reader
        
        # define and register a read function
        # the fmt defined the format key to handle the custom IO function
        # the types defined where the IO function will be applied, pre: preprocess, io: main io, post: postprocess
        @Reader.register(fmt='the/format/key', types='pre|io|post')
        def my_read_func(*arg, **kwargs)  # the args should meet the Reader demand
            ...
            
        # define and register a dump function
        def my_dump_func(*arg, **kwargs) # the args should meet the Dumper demand
            ...
        ---------------------------------- END ---------------------------------------
"""


# Define custom Exceptions
class IOEarlyStop(BaseException):
    """ monitor the situation that the IO should early stop and return None a the IO result """


# Define the IO function types
IOFuncPrefix = Literal['pre', 'io', 'post']
IOStream = Union[IOBase, str, bytes]


def _parse_lmp_data_script(script: str):
    """ Parse the LAMMPS data script to two dict, header and body"""
    # Define_body_title
    bt_name = (
        # atom-property sections
        'Atoms', 'Velocities', 'Masses', 'Ellipsoids', 'Lines', 'Triangles', 'Bodies',
        # molecular topology sections
        'Bonds', 'Angles', 'Dihedrals', 'Impropers',
        # force field sections
        'Pair Coeffs', 'PairIJ Coeffs', 'Bond Coeffs', 'Angle Coeffs', 'Dihedral Coeffs', 'Improper Coeffs',
        # class 2 force field sections
        'BondBond Coeffs', 'BondAngle Coeffs', 'MiddleBondTorsion Coeffs', 'EndBondTorsion Coeffs',
        'AngleTorsion Coeffs', 'AngleAngleTorsion Coeffs', 'BondBond13 Coeffs', 'AngleAngle Coeffs'
    )

    # Compile the body and header pattern
    header_title = re.compile(r'[a-z]+')
    header_int = re.compile(r'[0-9]+')
    header_float = re.compile(r'-?[0-9]+\.[0-9]*')

    lines = script.split('\n')
    body_split_point = [i for i, lin in enumerate(lines) if lin in bt_name] + [len(lines)]

    # Extract header info
    headers = {}
    for line in lines[1:body_split_point[0]]:
        line = line.strip()
        if line:
            ht = ' '.join(header_title.findall(line))
            hvs = line[:line.find(ht)].strip()  # header values
            header_values = []
            for hv in re.split(r'\s+', hvs):
                if header_int.fullmatch(hv):
                    header_values.append(int(hv))
                elif header_float.fullmatch(hv):
                    header_values.append(float(hv))
                else:
                    raise ValueError('the header line not match well')

            headers[ht] = header_values

    # Extract bodies info
    bodies = {}
    for sl_idx, el_idx in zip(body_split_point[:-1], body_split_point[1:]):
        bt = lines[sl_idx].strip()  # body title
        bc = [line.strip() for line in lines[sl_idx+1: el_idx] if line.strip()]  # body content
        bodies[bt] = bc

    return lines[0], headers, bodies


class Register:
    """
    Register the IO function for Dumper, Parser or so on
    """
    def __init__(self):
        # these dicts are container to store the custom io functions
        # the keys of the dict are serve as the have to get the mapped io functions(the values)
        self.pre_methods = {}
        self.io_methods = {}
        self.post_methods = {}

    def __repr__(self):
        return f"Register:\n" + \
               f"pre_method:\n" + \
               f"\n\t".join([n for n in self.pre_methods]) + "\n\n" + \
               f"io methods:\n" + \
               f"\n\t".join([n for n in self.io_methods]) + '\n\n' + \
               f"post methods:\n" + \
               f"\n\t".join([n for n in self.post_methods])

    def __call__(self, io_cls: type, fmt: str, prefix: IOFuncPrefix):
        """
        To register any function as a dumper or a postprocess to convert mol to formats
        Args:
            fmt:
            prefix:

        Returns:

        """

        def decorator(func: Callable):

            if prefix == 'pre':
                self.pre_methods[fmt] = func
            elif prefix == 'io':
                self.io_methods[fmt] = func
            elif prefix == 'post':
                self.post_methods[fmt] = func
            else:
                raise TypeError('the type of register is not supported')

            return func

        return decorator

    def pre(self, fmt: str):
        return self.pre_methods.get(fmt)

    def io(self, fmt: str):
        return self.io_methods.get(fmt)

    def post(self, fmt: str):
        return self.post_methods.get(fmt)


# Retrieve the IO class by it's format name
def retrieve_format(fmt: str = None):
    return _MoleculeIO.registered_format().get(fmt)


# Get all registered format name
def registered_format_name():
    return tuple(_MoleculeIO.registered_format().keys())


# TODO: deprecated in the later version
class _MoleculeIO(ABCMeta):
    """    Metaclass for registration of IO class format """
    _registered_format = {}

    def __new__(mcs, name, bases, namespace, **kwargs):
        # Get the format keywords
        fmt = namespace.get('format')(mcs)

        if not fmt:
            return super(_MoleculeIO, mcs).__new__(mcs, name, bases, namespace, **kwargs)
        elif not isinstance(fmt, str):
            raise TypeError('the defined format should be a string')
        elif fmt in mcs._registered_format:
            raise ValueError(f'the format {fmt} have been defined before')
        else:
            cls = super(_MoleculeIO, mcs).__new__(mcs, name, bases, namespace, **kwargs)
            mcs._registered_format[fmt] = cls
            return cls

    @classmethod
    def registered_format(mcs):
        return copy(mcs._registered_format)


class MetaIO(type):
    """
    Meta class to specify how to construct the IO class
    This Meta class is defined to register IO function conveniently.

    The IO functions are divided into three categories:
        - preprocess: do something before performing any of IO operation, with prefix '_pre'
        - io: performing the IO operation, with prefix '_io'
        - postprocess: do something after preforming IO operation, with prefix '_post'

    This Meta class offer two approach to defined and register the IO functions:
        - Define inside the IO class (IOClass)
        - Define outside the IO class and decorate the defined function by IOClass.register function

    To define inside the IOClass, one should name the IO function with the pattern:
        def _prefix_keys():
            ...
    where, the prefix is one of 'pre', 'io' or 'post'; the keys is the handle name to retrieve the
    IO functions.

    To define outside the IOClass, one should applied the class method register as the decorator of the
    IO functions, specified the prefix and the handle name as the decorator arguments, like:
        @IOClass.register(fmt='keys', types='prefix')
        def outside_io_func(*args, **kwargs):
            ...
    where the IOClass is one of Reader, Writer, Dumper, Parser or other custom IOClass, the 'key' and 'prefix'
    should be replace to the handle name and prefix you specified.
    """

    def __new__(mcs, name: str, bases: tuple, namespace: dict, **kwargs):
        """ If the subclasses contain methods with the prefix of '_pre', '_io' or '_post'
        they are seen as the IO function, that the preprocess, io or postprocess functions, respectively
        """
        _register = Register()

        for attr_name, attr in namespace.items():

            # Make sure the io function is a Callable obj
            if not isinstance(attr, Callable):
                continue

            split_names = attr_name.split('_')

            # the custom IO function should with prefix: '_pre', '_io' and '_post'
            # the handle keys of these function are follow the above prefix and separate be '_'
            # for example:
            #     def _pre_gjf(*args, **kwargs):
            #         ...
            # this is a preprocess IO function with a handle key: 'gjf' to retrieve the function.
            if len(split_names) <= 2:
                continue

            io_type = split_names[1]

            # Register the io functions:
            # if a define methods with the prefix '_pre', '_io' or '_post'
            # these methods are seen as preprocess, io or postprocess functions, respectively
            if io_type == 'pre':
                _register.pre_methods['_'.join(split_names[2:])] = attr
            if io_type == 'io':
                _register.io_methods['_'.join(split_names[2:])] = attr
            if io_type == 'post':
                _register.post_methods['_'.join(split_names[2:])] = attr

        namespace[f'_register'] = _register

        return type(name, bases, namespace, **kwargs)


class IOBase:
    """ The base IO class """
    # Initialize the register function, which is a callable obj embed in IO classes
    # When to register new IO function, apply the register function as decorator

    # _register = None

    def __init__(self, fmt: str, source: Union['ci.Molecule', IOStream], *args, **kwargs):
        """"""
        self.fmt = fmt
        self.src = source

        self.args = args
        self.kwargs = kwargs

        # override this methods to check the
        self.result = self._checks()

    def __call__(self):
        """ Call for the performing of IO """
        try:
            self._pre()
            # For dumper, the obj is Literal str or bytes obj
            # For parser, the obj is Molecule obj
            io_func = self._get_io()
            if io_func:  # If a custom io function have been defined, run custom functions
                obj = io_func(self)
            else:  # else get the general io function define in class
                obj = self._io()

            return self._post(obj)

        except IOEarlyStop:
            return None

    @abstractmethod
    def _checks(self) -> Dict[str, Any]:
        """
        This method should be override when definition of new IO class
        The purpose of this class is to check the regulation of initialized arguments.
        If not any arguments should be check, return None directly.
        """
        raise NotImplemented()

    def _get_pre(self) -> Callable:
        return self.register.pre(self.fmt)

    def _get_io(self) -> Callable:
        return self.register.io(self.fmt)

    def _get_post(self) -> Callable:
        return self.register.post(self.fmt)

    def _pre(self, *args, **kwargs):
        """ Regulate the method of preprocess """
        pre_func = self._get_pre()
        if pre_func:
            pre_func(self)

    @abstractmethod
    def _io(self, *args, **kwargs):
        """ Regulate the main io method """
        raise NotImplemented

    def _post(self, obj, *args, **kwargs):
        """ Regulate the method of postprocess """
        post_func = self._get_post()
        if post_func:
            return post_func(self, obj)
        else:
            return obj

    @property
    def register(self) -> Register:
        return getattr(self, f'_register')


class Dumper(IOBase, metaclass=MetaIO):
    """
    Dump the Molecule information into specific format.
    The output in general is the string or bytes
    """

    _pybel_fmt_convert = {
    }

    def _process_lmpdat_bonds(self, bond_contents: list):
        """"""
        uni_bonds = tuple(self.src.unique_bonds)
        bonds = self.src.bonds
        sep = re.compile(r'\s+')
        for i, bc in enumerate(bond_contents):
            split_bc = sep.split(bc)
            split_bc[1] = str(uni_bonds.index(bonds[i]) + 1)
            bond_contents[i] = '  '.join(split_bc)

        return bond_contents

    def _io(self):
        """ Performing the IO operation, convert the Molecule obj to Literal obj """
        # Try to dump by openbabel.pybel
        try:
            pb_mol = pybel.Molecule(self.src.ob_mol)
            return pb_mol.write(self._pybel_fmt_convert.get(self.fmt, self.fmt), **self.kwargs)

        except ValueError:
            print(IOError(f'the cheminfo.Molecule obj cannot dump to Literal'))
            return None

    def _checks(self) -> Dict[str, Any]:
        if not isinstance(self.src, ci.Molecule):
            raise TypeError(f'the dumped object should be hotpot.cheminfo.Molecule, instead of {type(self.src)}')

        return {}

    def _pre_cif(self):
        """
        pre-process for Molecule object to convert to cif file.
        if the hotpot object do not place in a Crystal, create a P1 compact Crystal for it
        """
        crystal = self.src.crystal()
        if not isinstance(crystal, ci.Crystal) or (
                np.logical_not(crystal.vector >= 0.).any() and np.logical_not(crystal.vector < 0.).any()
        ):
            self.src.compact_crystal(inplace=True)

        if self.src.crystal().space_group:
            self.src.crystal().space_group = 'P1'

    def _io_dpmd_sys(self):
        """ convert molecule information to numpy arrays """
        required_items = ['coord', 'type']
        check_atom_num = ['coord', 'force', 'charge']
        share_same_configures = ['type', 'coord', 'energy', 'force', 'charge', 'virial']
        need_reshape = ['coord', 'force']

        conf_num = len(self.src.all_coordinates)
        crystal = self.src.crystal()
        if crystal:
            box = self.src.crystal().vector  # angstrom
            is_periodic = True
        else:
            box = np.zeros((3, 3))
            for i in range(3):
                box[i, i] = 100.
            is_periodic = False
        box = box.reshape(-1, 9).repeat(conf_num, axis=0)

        data = {
            'type': np.array(self.src.atomic_numbers).reshape(1, -1).repeat(conf_num, axis=0),
            'type_map': ['-'] + list(ci.periodic_table.keys()),
            'nopbc': not is_periodic,
            'coord': self.src.all_coordinates,  # angstrom,
            'box': box,
            'energy': self.src.all_energy,  # eV
            'force': self.src.all_forces,  # Hartree/Bohr,
            'charge': self.src.all_atom_charges,  # q
            'virial': None,
            'atom_ener': None,
            'atom_pref': None,
            'dipole': None,
            'atom_dipole': None,
            'polarizability': None,
            'atomic_polarizability': None
        }

        for name in required_items:
            if data.get(name) is None:
                raise ValueError('the required composition to make the dpmd system is incomplete!')

        # Check whether the number of conformers are matching among data
        if any(len(data[n]) != conf_num for n in share_same_configures if data[n] is not None):
            raise ValueError('the number of conformers is not match')

        # Check whether the number of atoms in data are matching to the molecular atoms
        if any(data[n].shape[1] != self.src.atom_num for n in check_atom_num if data[n] is not None):
            raise ValueError('the number of atoms is not matching the number of atom is the molecule')

        for name in need_reshape:
            item = data.get(name)
            if isinstance(item, np.ndarray):
                shape = item.shape

                assert len(shape) == 3

                data[name] = item.reshape((shape[0], shape[1]*shape[2]))

        return data

    def _io_lmpmol(self):
        """
        write a molecule script
        default values: coordinates, velocities, atom IDs and types
        additional attributes for atomic: Bonds
        additional attributes for full: Bonds + molecular + charge
        """

        def bonds(m):
            """ Add bond body """
            bond_str = 'Bonds' + '\n\n'  # bond body title

            # the formula of bond_type key: atom1[bond_type]atom2
            uni_bonds = tuple(m.unique_bonds)  # store bonds type
            for j, bond in enumerate(m.bonds, 1):
                a1_idx = atoms_list.index(bond.atom1) + 1
                a2_idx = atoms_list.index(bond.atom2) + 1

                bt_id = uni_bonds.index(bond) + 1

                bond_str += f'{j} {bt_id} {a1_idx} {a2_idx}\n'

            bond_str += '\n'

            return bond_str

        def charge():
            """ Retrieve atom charge information """
            charge_str = '\n' + 'Charges' + '\n\n'

            for ic, a in enumerate(atoms_list, 1):  # ID of charge, atom
                if isinstance(a, ci.Atom):
                    charge_str += f'{ic} {a.partial_charge}\n'
                else:
                    assert isinstance(a, ci.PseudoAtom)
                    charge_str += f'{ic} {a.charge}\n'

            charge_str += '\n'

            return charge_str

        mol = self.src
        kwargs = self.kwargs  # keywords arguments

        # default values: coordinates, velocities, atom IDs and types;
        # additional attributes for atomic: None;
        # additional attributes for full: molecular + charge
        atom_style = kwargs.get('atom_style', 'atomic')   # default atom_style is atomic
        mol_name = kwargs.get('mol_name', mol.smiles)

        # combine real atoms with pseudo atoms in a list
        atoms_list = []
        for m_a in mol.atoms:
            atoms_list.append(m_a)

        if mol.pseudo_atoms:   # determine if there are pseudo_atoms
            for pse_a in mol.pseudo_atoms:
                atoms_list.append(pse_a)

        # title information
        title = f"Create by hotpot package, convert from {mol_name}"
        script = title + '\n\n'   # write the molecular script for lammps

        # TODO: some header information missing
        # Header partition
        # add atom header
        num_atoms = len(atoms_list)
        num_atoms_str = f'{num_atoms}  atoms'
        script += num_atoms_str + '\n'

        # add bond header
        num_bonds = len(mol.bonds)
        num_bonds_str = f'{num_bonds}  bonds'
        script += num_bonds_str + '\n'

        # Add new blank line to end the header partition
        script += '\n'

        # Body partition
        # Coords body
        script += 'Coords' + '\n\n'
        for i, atom in enumerate(atoms_list, 1):
            script += f'{i}' + '  ' + '  '.join(map(str, atom.coordinates)) + '\n'
        script += '\n'

        # Types body
        script += 'Types' + '\n\n'

        dict_types = {}
        for i, atom in enumerate(atoms_list, 1):
            atom_type = dict_types.setdefault(atom.symbol, len(dict_types)+1)
            script += f'{i} {atom_type}  # {atom.symbol}\n'

        script += '\n'

        # additional attributes
        # to atomic style, only basis information (ID，Coords, types, velocitier)
        if atom_style == 'atomic':
            if num_bonds:
                script += bonds(mol)

        # to full style, basis information + molecular + charge
        elif atom_style == 'full':
            if num_bonds:
                script += bonds(mol)
            script += charge()

        return script

    def _post_dpmd_sys(self, data: Dict[str, Union[List, np.ndarray, None]]):
        """"""
        mapping_items = ['type', 'type_map']
        path_save = self.kwargs.get('path_save')
        if path_save:
            if isinstance(path_save, str):
                path_save = Path(path_save)
            if path_save.is_file():
                raise NotADirectoryError('the given path_save to dpmd_sys should be an empty dir, not a existed file')

            if not path_save.exists():
                path_save.mkdir()

            if [p for p in path_save.glob('*')]:
                raise IOError('the given path_save should be a empty dir, files or dirs have exist in the dir')

            dir_npy = path_save.joinpath('set.000')
            dir_npy.mkdir()

            for name, value in data.items():
                if name == 'nopbc' and value:
                    with open(path_save.joinpath('nopbc'), 'w') as writer:
                        writer.write('')

                elif name in mapping_items:
                    with open(path_save.joinpath(f'{name}.raw'), 'w') as writer:
                        writer.write('\n'.join(map(str, value)))

                else:
                    if value is not None:
                        if not isinstance(value, np.ndarray):
                            raise ValueError(f'the data to set.*** must be np.ndarray, instead of {type(value)}')

                        np.save(dir_npy.joinpath(f'{name}.npy'), value)

        return data

    def _post_gjf(self, script):
        """ postprocess the dumped Gaussian 16 .gjf script to add the link0 and route context """
        # To count the insert lines
        inserted_lines = 0

        # separate keyword arguments:
        link0 = self.kwargs['link0']
        route = self.kwargs['route']
        custom_charge = self.kwargs.get('charge')
        custom_spin = self.kwargs.get('spin')

        lines = script.splitlines()

        # Write link0 command
        if isinstance(link0, str):
            lines[0] = f'%{link0}'
        elif isinstance(link0, list):
            for i, stc in enumerate(link0):  # stc=sentence
                assert isinstance(stc, str)
                if not i:  # For the first line of link0, replace the the original line in raw script
                    lines[0] = f'%{stc}'
                else:  # For the other lines, insert into after the 1st line
                    inserted_lines += 1
                    lines.insert(inserted_lines, f'%{stc}')
        else:
            raise TypeError('the link0 should be string or list of string')

        # Write route command
        if isinstance(route, str):
            lines[1+inserted_lines] = f'# {route}'
        elif isinstance(route, list):
            for i, stc in enumerate(route):
                assert isinstance(stc, str)
                if not i:  # For the first line of link0, replace the the original line in raw script
                    lines[1+inserted_lines] = f'#{stc}'
                else:  # For the other lines, insert into after the original route line.
                    inserted_lines += 1
                    lines.insert(inserted_lines+1, f'%{stc}')
        else:
            raise TypeError('the route should be string or list of string')

        charge, spin = lines[5+inserted_lines].split()
        if custom_charge:
            charge = str(custom_charge)
        if custom_spin:
            spin = str(custom_spin)

        lines[5+inserted_lines] = f'{charge} {spin}'

        script = '\n'.join(lines)

        # End black line
        script += '\n\n'

        return script

    def _post_lmpdat(self, script: str):
        """ post-process for LAMMPS data file """

        title, headers, bodies = _parse_lmp_data_script(script)
        script = title + '\n'

        for ht, hvs in headers.items():
            if ht == 'bond types':  # header title, header values
                hvs[0] = len(self.src.unique_bonds)

            script += ' '.join(map(str, hvs)) + ' ' + ht + '\n'

        script += '\n' * 3

        for bt, bcs in bodies.items():  # body title, body contents
            if bt == 'Bonds':
                bcs = self._process_lmpdat_bonds(bcs)

            if bcs:  # if the body contents exist
                script += bt + '\n' * 2
                script += '\n'.join(bcs)
                script += '\n' * 3

        return script


class Parser(IOBase, metaclass=MetaIO):
    """ Parse the str or bytes obj to Molecule obj """
    _pybel_fmt_convert = {
        'g16log': 'g16'
    }

    def _open_source_to_string_lines(self, *which_allowed: str, output_type: Literal['lines', 'script'] = 'lines'):
        """
        Open the source file to string lines
        Args:
            which_allowed: which types of source are allowed to process to string lines

        Returns:
            (List of string|string)
        """
        src_type = self.result.get('src_type')
        if src_type not in which_allowed:
            raise RuntimeError(f'the source type {type(self.src)} have not been supported')
        else:
            if src_type == 'str':
                script = self.src

            elif src_type == 'path':
                with open(self.src) as file:
                    try:
                        script = file.read()
                    # If the file pointed by the path is not a text file
                    # such as a bytes file
                    except UnicodeDecodeError:
                        raise IOEarlyStop()

            elif src_type == 'IOString':
                script = self.src.read()

            else:
                raise RuntimeError(f'the source type {type(self.src)} have not been supported')

            if output_type == 'lines':
                return script.split('\n')
            elif output_type == 'script':
                return script
            else:
                raise ValueError('the arg output_type given a wrong values, lines or script allow only')

    def _checks(self) -> Dict[str, Any]:
        if not isinstance(self.src, (IOBase, str, bytes, PathLike)):
            raise TypeError(f'the parsed object should be IOBase, str or bytes, instead of {type(self.src)}')

        if isinstance(self.src, str):
            if os.path.exists(self.src):
                return {'src_type': 'path'}
            else:
                return {'src_type': 'str'}

        if isinstance(self.src, PathLike):
            return {'src_type': 'path'}

        if isinstance(self.src, bytes):
            return {'src_type': 'bytes'}
        if isinstance(self.src, io.StringIO):
            return {'src_type': 'StringIO'}
        if isinstance(self.src, io.BytesIO):
            return {'src_type': 'BytesIO'}
        if isinstance(self.src, io.FileIO):
            return {'src_type': 'FileIO'}
        print(f'the get source type is {type(self.src)}')
        return {'src_type': type(self.src)}

    def _ob_io(self):
        """ IO by openbabel.pybel """
        # Get the source type name
        src_type = self.result.get('src_type')
        try:
            if src_type == 'str':
                pybel_mol = pybel.readstring(self._pybel_fmt_convert.get(self.fmt, self.fmt), self.src)
            elif src_type == 'path':
                pybel_mol = next(pybel.readfile(self._pybel_fmt_convert.get(self.fmt, self.fmt), str(self.src)))
            elif src_type == 'IOString':
                pybel_mol = pybel.readstring(self._pybel_fmt_convert.get(self.fmt, self.fmt), self.src.read())
            else:
                raise RuntimeError(f'the source type {type(self.src)} have not been supported')

            obj = ci.Molecule(pybel_mol.OBMol)

        except RuntimeError:
            obj = None

        return obj

    def _cclib_io(self, obj):
        """ IO by cclib package """
        src_type = self.result.get('src_type')

        try:
            if src_type == 'str':
                data = cclib.ccopen(io.StringIO(self.src)).parse()
            elif src_type == 'path':
                data = cclib.ccopen(self.src).parse()
            elif src_type == 'IOString':
                data = cclib.ccopen(self.src).parse()
            else:
                raise RuntimeError(f'the source type {type(self.src)} have not been supported in cclib')

        except (RuntimeError, AttributeError):
            data = None

        if data:
            if not obj:
                # if get information about the atoms species
                if hasattr(data, 'atomnos'):
                    atoms_attrs = [{'atomic_number': an} for an in getattr(data, 'atomnos')]
                    obj = ci.Molecule(atoms=atoms_attrs)
                else:
                    print(IOError(f'the parsing of {self.src} is not successful!'))
                    return obj  # Return None

            # if get information about the coordination collections
            if hasattr(data, 'atomcoords'):
                obj.set(all_coordinates=getattr(data, 'atomcoords'))

            # if get information about the energy (SCF energies) vector
            if hasattr(data, 'scfenergies'):
                obj.set(all_energy=getattr(data, 'scfenergies'))

        return obj

    def _io(self, *args, **kwargs):
        """ Standard IO process """
        # Try parse the log file by openbabel.pybel file firstly
        return self._ob_io()

    # Start to the prefix IO functions

    # preprocess for g16log file
    # This preprocess is used to judge whether a Error happened when perform g16 calculate
    def _pre_g16log(self):
        """ g16log preprocess to judge whether some Error happened """
        def is_convergence_failure():
            if 'Convergence failure -- run terminated.' in script:
                return True
            return False

        def is_hessian_no_longer_linear_valid():
            march_pattern = re.compile(r'Error termination via Lnk1e in (/.+)*/l103\.exe')

            if any (march_pattern.match(line.strip()) for line in script.splitlines()[-5:]):
                return True
            return False

        script = self._open_source_to_string_lines('str', 'path', "IOString", output_type='script')

        # Check whether a failure have happened when calculation.
        if is_convergence_failure():
            raise IOEarlyStop('Gaussian16 SCF cannot convergence!')
        if is_hessian_no_longer_linear_valid():
            raise IOEarlyStop('Gaussian16 Hessian no longer linear valid')

    # Parse the XYZ file
    def _io_xyz(self):
        """ Parse the XYZ file """
        src_type = self.result['src_type']
        if src_type != 'path':
            return self._io()
        else:
            from ase import io
            from openbabel import openbabel as ob
            data_generator = io.iread(self.src)

            # atomic_numbers, coordinates, cell_params
            atomic_numbers, all_coordinates, cell_matrix = [], [], None
            for data in data_generator:
                atomic_numbers.append(data.numbers)
                all_coordinates.append(data.positions)
                if cell_matrix is None:
                    cell_matrix = data.cell.array

            atomic_numbers = np.stack(atomic_numbers)
            all_coordinates = np.stack(all_coordinates)

            number_min = atomic_numbers.min(axis=0)
            number_max = atomic_numbers.max(axis=0)

            # the values in same columns should be same.
            assert all(number_max == number_min)

            obj = ci.Molecule()
            obj.quick_build_atoms(number_min)

            obj.set(all_coordinates=all_coordinates)
            obj.set(crystal=cell_matrix)
            obj.configure_select(0)

            return obj

    # postprocess for g16log file
    def _post_g16log(self, obj: 'ci.Molecule'):
        """
        post process for g16log format, to extract:
            1) Mulliken charge
            2) Spin densities
        """
        def extract_charges_spin():
            """ Extract charges and spin information from g16.log file """
            # Get the line index of Mulliken charges
            head_lines = [i for i, line in enumerate(lines) if line.strip() == 'Mulliken charges and spin densities:']
            
            if not head_lines:
                raise IOEarlyStop
            elif len(head_lines) == obj.configure_number + 1:
                head_lines = head_lines[1:]

            # Extract the Mulliken charge and spin densities
            charges, spin_densities = [], []  # changes(cgs) spin_densities(sds)
            for i in head_lines:
                # Enhance inspection
                col1, col2 = lines[i + 1].strip().split()
                assert col1 == '1' and col2 == '2'

                HEAD_LINES_NUM = 2
                cg, sd = [], []  # change, spin_density

                while True:
                    split_line = lines[i + HEAD_LINES_NUM].strip().split()
                    if len(split_line) != 4:
                        break
                    else:
                        row, syb, c, s = split_line  # row number, symbol, charges, spin density

                    try:
                        row, c, s = int(row), float(c), float(s)
                        # check the sheet row number
                        if row != HEAD_LINES_NUM - 1:
                            break

                    # Inspect the types of values
                    except ValueError:
                        break

                    # record the charge and spin density
                    cg.append(c)
                    sd.append(s)
                    HEAD_LINES_NUM += 1

                # store the extracted
                if cg and sd:
                    if len(cg) == len(sd) == len(obj.atoms):
                        charges.append(cg)
                        spin_densities.append(sd)
                    else:
                        raise ValueError('the number of charges do not match to the number of atoms')
                else:
                    raise ValueError('get a empty charge and spin list, check the input!!')

            obj.set(all_atom_charges=np.array(charges))
            obj.set(all_atom_spin_densities=np.array(spin_densities))
                
        def extract_force_matrix():
            # Define the format of force sheet
            # the Force sheet like this:
            #  -------------------------------------------------------------------
            #  Center     Atomic                   Forces (Hartrees/Bohr)
            #  Number     Number              X              Y              Z
            #  -------------------------------------------------------------------
            #       1        8           0.039901671    0.000402574    0.014942530
            #       2        8           0.017381613    0.001609531    0.006381231
            #       3        6          -0.092853735   -0.025654844   -0.005885898
            #       4        6           0.067801154    0.024130172   -0.022794721
            #       5        8          -0.023702905    0.005486251   -0.004938175
            #       6        8          -0.006359715   -0.008543465    0.010350815
            #       7       55          -0.002168084    0.002569781    0.001944217
            #  -------------------------------------------------------------------
            force_head1 = re.compile(r'\s*Center\s+Atomic\s+Forces\s\(Hartrees/Bohr\)\s*')
            force_head2 = re.compile(r'\s*Number\s+Number\s+X\s+Y\s+Z\s*')
            sheet_line = re.compile(r'\s*----+\s*')

            HEAD_LINES_NUM = 3  # the offset line to write the header

            head_lines = [i for i, line in enumerate(lines) if force_head1.match(line)]

            all_forces = []
            for i in head_lines:
                # enhance the inspection of Force sheet head
                assert force_head2.match(lines[i + 1])
                assert sheet_line.match(lines[i + 2])

                rows = 0
                forces = []
                while True:

                    if sheet_line.match(lines[i + HEAD_LINES_NUM + rows]):
                        if len(forces) == obj.atom_num:
                            all_forces.append(forces)
                            break
                        else:
                            raise ValueError('the number of force vector do not match the number of atoms')

                    ac, an, x, y, z = map(
                        lambda v: int(v[1]) if v[0] < 2 else float(v[1]),
                        enumerate(lines[i + HEAD_LINES_NUM + rows].split())
                    )

                    # Enhance the inspection
                    assert ac == rows + 1
                    if obj.atoms[rows].atomic_number != an:
                        raise ValueError('the atomic number do not match')

                    forces.append([x, y, z])

                    rows += 1

            try:
                obj.set(all_forces=np.array(all_forces))
            except ValueError:
                return

        obj = self._cclib_io(obj)  # Try to supplementary Molecule data by cclib

        lines = self._open_source_to_string_lines('str', 'path', 'IOString')

        try:  # TODO: For now, this is the case, the spin densities may lost in some case  # the units is Hartree/Bohr
            extract_charges_spin()
            extract_force_matrix()
        except IndexError:
            raise IOEarlyStop

        # assign the first configure for the molecule
        obj.configure_select(0)

        return obj





"""
python v3.7.9
@Project: hotpot
@File   : __init__.py.py
@Author : Zhiyuan Zhang
@Date   : 2023/3/14
@Time   : 4:06
"""
import os
import sys

sys.path.append(os.path.abspath(os.path.dirname(__file__)))

src_root = os.path.abspath(os.path.dirname(__file__))
data_root = os.path.abspath(os.path.join(src_root, 'data'))

from cheminfo import Molecule
from bundle import MolBundle



"""
python v3.7.9
@Project: hotpot
@File   : test_chemif.py
@Author : Zhiyuan Zhang
@Date   : 2023/3/15
@Time   : 3:47
"""
import hotpot.cheminfo as ci

def mol2_read():
    path_mol2 = 'examples/struct/mol.mol2'
    mol = ci.Molecule.read_from(path_mol2)

    return mol


def dump_gjf():
    path_mol2 = 'examples/struct/mol.mol2'
    mol = ci.Molecule.read_from(path_mol2)

    script = mol.dump('gjf',
                      link0='CPU=0-48',
                      route='M062X/Def2SVP opt(MaxCyc=10)/freq SCRF pop(Always)',
                      )

    print(script)


def parse_g16log():
    path_mol1 = 'examples/struct/0.log'
    path_mol2 = 'examples/struct/222.log'
    path_mol3 = 'examples/struct/Cs-VOHSAM-5.mol2'

    mol1 = ci.Molecule.read_from(path_mol1, 'g16log')
    mol2 = ci.Molecule.read_from(path_mol2, 'g16log')
    mol3 = ci.Molecule.read_from(path_mol3)

    return mol1, mol2, mol3


def perturb_cif():
    path_cif = 'examples/struct/aCarbon.cif'
    mol = ci.Molecule.read_from(path_cif, 'cif')
    gen = mol.perturb_mol_lattice(mol_distance=0.05, max_generate_num=2)

    for i, gen_mol in enumerate(gen):
        gen_mol.writefile('cif', f'output/gen_cif/aCarbon_{i}.cif')


def gen_pairs():
    from openbabel import openbabel as ob
    mol = ci.Molecule.read_from('c1ccc(C(=O)O)c(O)c1CCCC', 'smi')
    mol2 = ci.Molecule.read_from('c1ccc(C(=O)O)c(O)c1CCCC', 'smi')
    mol.build_conformer('UFF')
    mol2.build_conformer()
    mol.normalize_labels()
    mol2.normalize_labels()
    g = mol.generate_metal_ligand_pair('Sr')

    # for a1, a2 in zip(mol.atoms, mol2.atoms):
    #     print(f'{a1.label}: {a1.partial_charge}---{a2.label}: {a2.partial_charge}')

    ps = [p for i, p in enumerate(g)]
    ps[1].remove_atoms('C1', 'C2')
    ps[1].writefile('mol2', f'/home/zz1/g01.mol2')


if __name__ == '__main__':
    mol = ci.Molecule.read_from('/home/zz1/ZONTOO.cif')
    # mol.build_conformer()
    # mol.remove_solvents()
    # mol.writefile('mol2', '/home/zz1/self.mol2')
    ligands = mol.retrieve_ligands()
    for ligand in ligands:
        ligand.add_hydrogens()
        ligand.localed_optimize()
    pairs = [p for l in ligands for p in l.generate_metal_ligand_pair('Sr')]



"""
python v3.9.0
@Project: hotpot
@File   : cc
@Auther : Zhiyuan Zhang
@Data   : 2023/6/4
@Time   : 21:05

This module is to perform `Task` about Coordination Chemistry (cc)
"""
from ..cheminfo import Molecule

class Ligand(Molecule):
    """ Represent a ligand """


class MetalLigandPair(Molecule):
    """ Represent a metal-organic pair """




"""
python v3.7.9
@Project: hotpot
@File   : utils.py
@Author : Zhiyuan Zhang
@Date   : 2023/3/29
@Time   : 0:30

Notes:
    Private common functions
"""
from os import PathLike
from pathlib import Path
from typing import *


class PathNotExistError(Exception):
    """ Raise when can't find a file or dir """


def check_path(
        path: Optional[Union[str, PathLike]],
        none_allowed: Optional[bool] = True,
        check_exist: Optional[bool] = False,
        mkdir: Optional[bool] = False,
        file_or_dir: Optional[Literal['file', 'dir']] = None
) -> Union[NoReturn, Path]:
    """
    Check whether the given path is valid and process str path to a Path object.

    Args:
        path (Optional[Union[str, PathLike]]): The path to be checked.
        none_allowed (Optional[bool], default=True): Whether to allow the path to be None.
        check_exist (Optional[bool], default=False): Whether to check if the path exists.
        mkdir (Optional[bool], default=False): Whether to create the directory if it doesn't exist.
        file_or_dir (Optional[Literal['file', 'dir']], default=None): Whether to check if the path is a file or directory.

    Returns:
        Union[None, Path]: The processed path as a Path object or None if allowed and given as None.

    Raises:
        ValueError: If the path is None and not allowed, or if `mkdir=True` and `file_or_dir='file'`.
        TypeError: If the given path is neither a str nor a PathLike object.
        PathNotExistError: If the path doesn't exist and check_exist is True or mkdir is False.
        IsADirectoryError: If the path is a directory and file_or_dir is 'file'.
        NotADirectoryError: If the path is a file and file_or_dir is 'dir'
    """
    # If get the None, keep it or raise!
    if not path:
        if none_allowed:
            return path
        else:
            raise ValueError("the path shouldn't to be None")

    # Check the given path type and transform the str path to Path
    if isinstance(path, str):
        path = Path(path)
    elif not isinstance(path, PathLike):
        raise TypeError(f'the given path should be str or PathLick, instead of{type(path)}')

    is_exist = path.exists() if (check_exist or mkdir or file_or_dir) else None

    if is_exist is False:
        if file_or_dir == 'file' and mkdir:
            raise ValueError("the mkdir=True and file_or_dir='file' can't to be set simultaneously!")

        if mkdir:
            path.mkdir()
        else:
            raise PathNotExistError(f'the path {str(path)} not exist!')

    if file_or_dir == 'file' and not path.is_file():
        raise IsADirectoryError(f'{str(path)} is a directory!')
    if file_or_dir == 'dir' and not path.is_dir():
        raise NotADirectoryError(f'{str(path)} is a file!')

    return path



"""
python v3.7.9
@Project: hotpot
@File   : test_bundle.py
@Author : Zhiyuan Zhang
@Date   : 2023/3/19
@Time   : 21:25
"""
from hotpot import MolBundle
import json
import hotpot.cheminfo as ci


if __name__ == '__main__':
    dir_log = '/home/zz1/proj/gauss/new/log'
    md = MolBundle.read_from_dir('g16log', dir_log, match_pattern='*/*.log', num_proc=16)
    mmd = md.merge_conformers()
    mamd = mmd.merge_atoms_same_mols()

    count = 0
    list_error = []
    for m in md:
        try:
            data = m.dump('dpmd_sys')
        except ValueError:
            count += 1
            list_error.append(m.identifier)
            print(m.configure_number, len(m.all_energy), len(m.all_forces), len(m.all_atom_spin_densities))

    # list_error = json.load(open('/home/zz1/error_file.json'))
    # for p_e in list_error:
    #     m = ci.Molecule.read_from(p_e, 'g16log')
    #     if m:
    #         print(m.configure_number, len(m.all_energy), len(m.all_forces), len(m.all_atom_spin_densities))

    dm = mmd[0].to_dpmd_sys('/home/zz1/proj/dpmd/std')
    dma = mamd[0].to_dpmd_sys('/home/zz1/proj/dpmd/att', 'att')



"""
python v3.7.9
@Project: hotpot
@File   : test_lmp.py
@Author : Zhiyuan Zhang
@Date   : 2023/4/24
@Time   : 21:42
"""
import random

import numpy as np

import hotpot.cheminfo as ci


def melt():
    mol = ci.Molecule.read_from('examples/struct/aCarbon.xyz')

    mol.lmp_setup()

    mol.lmp.command('units metal')
    mol.lmp.command('dimension 3')
    mol.lmp.command('atom_style full')
    mol.lmp.command('\n')
    mol.lmp.command('# ----------------- Atom Definition Section -----------------')
    mol.lmp.command('\n')
    mol.lmp.read_main_data()

    mol.lmp.command('pair_style tersoff')
    mol.lmp.command('pair_coeff * * /home/zz1/qyq/SiC.tersoff C')

    mol.lmp.command('fix melt all nvt temp 300 10000 100')
    mol.lmp.command('fix melt all nvt temp 10000 10000 1000.0')
    mol.lmp.command('timestep 0.00007')

    mol.lmp.command('thermo_style    custom step temp pe etotal press vol density')
    mol.lmp.command('thermo 1000')

    mol.lmp.command('dump dmelt all xyz 1000 /home/zz1/qyq/dmelt.xyz')
    mol.lmp.command('dump_modify dmelt element C')
    mol.lmp.command('restart 10000 /home/zz1/qyq/dmelt.xyz')

    mol.lmp.command('run 1400000')


def quench():
    mol = ci.Molecule.read_from('/home/zz1/qyq/dmelt.xyz')
    mol.lmp_setup()
    mol.lmp.commands_string(
        """
        units metal
        dimension 3
        atom_style full
        """
    )
    mol.lmp.read_main_data()
    mol.lmp.command("pair_style tersoff")
    mol.lmp.command("pair_coeff * * /home/zz1/qyq/SiC.tersoff C""")

    mol.lmp.command('thermo_style    custom step temp pe etotal press vol density')
    mol.lmp.command('thermo          10000')
    mol.lmp.command('timestep 0.00007')

    mol.lmp.command('dump xyz all xyz 1000 /home/zz1/qyq/quench.xyz')
    mol.lmp.command('dump_modify xyz element C')

    mol.lmp.command("fix quench all  nvt temp 10000.0 3000.0 7.0")
    mol.lmp.command('run 100000')
    mol.lmp.command('fix quench all  nvt temp 3000.0 3000.0 7.0')
    mol.lmp.command('run 100000')
    mol.lmp.command('fix quench all  nvt temp 3000.0 300.0 7.0')
    mol.lmp.command('run 100000')
    mol.lmp.command('fix quench all  nvt temp 3000.0 300.0 7.0')
    mol.lmp.command('run 100000')
    mol.lmp.command('fix quench all  nvt temp 300.0 300.0 7.0')
    mol.lmp.command('run 100000')


def melt_quench(
        m: ci.Molecule,
        path_dump_to: str,
        origin_temp: float = 298.15,
        relax_temp: float = 3500,
        melt_temp: float = 10000,
        time_step: float = 0.00007
):
    m.lmp_setup()

    # initialization
    m.lmp.commands_string(
        """
        units metal
        dimension 3
        atom_style full
        """
    )

    # Read molecule into LAMMPS
    m.lmp.read_main_data()

    # Configure the force field
    m.lmp.command("pair_style tersoff")
    m.lmp.command("pair_coeff * * /home/zz1/qyq/SiC.tersoff C""")

    # Specify the thermodynamical output to screen
    m.lmp.command('thermo_style    custom step temp pe etotal press vol density')
    m.lmp.command('thermo          1000')

    # the step interval of integral
    m.lmp.command(f'timestep {time_step}')

    # Specify the dump configuration
    m.lmp.command(f'dump xyz all xyz 100 {path_dump_to}')
    m.lmp.command('dump_modify xyz element C')

    # Initialize the temperature for system
    m.lmp.command(f'velocity all create {origin_temp} {random.randint(100000, 999999)}')

    # temperature rise up
    m.lmp.command(f'fix 0 all nvt temp {origin_temp} {melt_temp} 0.7')
    m.lmp.command(f'run 10000')

    m.lmp.command(f'fix 0 all nvt temp {melt_temp} {melt_temp} 1.4')
    m.lmp.command(f'run 20000')

    # Relax
    m.lmp.command('thermo          250')
    m.lmp.command(f'fix 0 all nvt temp {relax_temp} {relax_temp} 1000.0')
    while m.lmp.eval('temp') > relax_temp*1.05:
        m.lmp.command(
            f'velocity all scale {relax_temp}'
        )
        m.lmp.command(f'run 2000')

    m.lmp.command('thermo          1000')
    m.lmp.command(f'run 20000')

    # Quench
    m.lmp.command('thermo          250')
    m.lmp.command(f'fix 0 all nvt temp {origin_temp} {origin_temp} 1000.0')
    while m.lmp.eval('temp') > 400:
        m.lmp.command(
            f'velocity all create {(m.lmp.eval("temp")-origin_temp) / 2 + origin_temp}'
        )
        m.lmp.command(f'run 2000')

    return m


def gcmc():
    """"""
    p = 52.5573
    iodine = ci.Molecule()
    iodine.add_pseudo_atom('iodine', 253.808946, (0., 0., 0.))

    mol = ci.Molecule.read_from('/home/zz1/qyq/mq.cif')
    mol.lmp_setup()

    mol.lmp('units real')
    mol.lmp('dimension 3')
    mol.lmp('atom_style full')

    mol.lmp('read_data /home/zz1/qyq/main.data group frame extra/atom/types 1')
    mol.lmp('mass 2 253.808946')
    # mol.lmp.read_main_data(group='frame')

    # write molecule read file
    # iodine.writefile('lmpmol', '/home/zz1/qyq/lmpmol')
    mol.lmp('molecule iodine /home/zz1/qyq/lmpmol toff 1')
    mol.lmp('group Igas empty')

    mol.lmp('pair_style lj/cut 5.5')
    mol.lmp(f'pair_coeff 1 1 {52.83*0.001987} 3.43')
    mol.lmp(f'pair_coeff 2 2 {550.0*0.001987} 4.982')

    mol.lmp('fix stand frame setforce 0.0 0.0 0.0')

    mol.lmp(
        'fix gcmc Igas gcmc 1 $((20+abs(count(Igas)-20))/2) $((20+abs(count(Igas)-20))/2)'
        ' 0 354568 298.15 100 10 mol iodine pressure 52.5573 tfac_insert 15'
    )

    mol.lmp(
        f'fix out all print 1 "$(step),$(count(Igas)),$(mass(Igas)),$(mass(Igas)/mass(frame))" '
        f'file /home/zz1/qyq/I2.csv '
        'title "step,I2_count,I2_mass,uptake(g/g)" screen no'
    )
    mol.lmp('variable uptake equal mass(Igas)/mass(frame)')
    mol.lmp('fix ave all ave/time 1 50 50 v_uptake file /home/zz1/qyq/ave')

    mol.lmp('thermo_style    custom step temp pe etotal press vol density')
    mol.lmp('thermo          1000')
    mol.lmp('compute_modify thermo_temp dynamic/dof yes')

    mol.lmp(f'dump mq all xyz 100 /home/zz1/qyq/I2.xyz')
    mol.lmp.command(f'dump_modify mq element C, I')

    mol.lmp('timestep 0.0001')
    mol.lmp('run 15000')

    return mol


if __name__ == '__main__':
    from hotpot.tanks.lmp.gcmc import LjGCMC
    import os
    import tqdm
    from time import time as t
    frame = ci.Molecule.read_from('/home/zz1/qyq/mq.cif')
    # frame.build_bonds()
    # # ff = frame + frame
    #
    # # script = frame.dump('lmpdat')
    # #
    # # g1 = ci.Molecule.read_from('O=O', 'smi')
    # # g1.build_conformer()
    # g2 = ci.Molecule.read_from('O=C=O', 'smi')
    # g2.build_conformer()
    #
    # g2 + frame
    # g3 = ci.Molecule.read_from('N#N', 'smi')
    # g3.build_conformer()
    # g4 = ci.Molecule.read_from('O', 'smi')
    # g4.build_conformer()
    #
    # work_dir = f'/home/zz1/qyq/g1g2g3'
    # gc = LjGCMC(frame, 'UFF/LJ.json', g2, g4, work_dir=work_dir, T=298.15, P=0.00001)
    # gc.run()



"""
python v3.9.0
@Project: hotpot
@File   : load_chem_lib
@Auther : Zhiyuan Zhang
@Data   : 2023/6/8
@Time   : 3:27

This module is used to lazily load the chemical information database when other modules need it.
"""
from pathlib import Path

class Library:
    """ the Main class to load and save chemical information lazily """

    _lib = {}  # library for chemical books

    def __init__(self):
        self._books = {}

    @property
    def book_list(self):
        return list(self._lib.keys())

    @classmethod
    def register(cls, book_class: type):
        """ sign up the chemical books """
        cls._lib[book_class.__name__] = book_class

    def get(self, book_name: str):
        return self._books.setdefault(book_name, self._lib[book_name]())


class ChemicalBook:
    """ The base class for all chemical books """

@Library.register
class Solvents(ChemicalBook):
    """ the ChemicalBook to store common solvents """
    def __init__(self):
        dir_solvents = Path(hp.data_root).joinpath('solvents')
        self._solvents = [hp.Molecule.read_from(p) for p in dir_solvents.glob('*.mol2')]
        self._sols_smi = [m.smiles for m in self._solvents]

    def __iter__(self):
        return iter(self._solvents)

    def __getitem__(self, item):
        return self._solvents[item]

    def __repr__(self):
        return f'SolventsBook({len(self._solvents)})'

    def is_solvent(self, mol: 'hp.Molecule'):
        """ to judge whether a molecule is a solvent """
        return mol.smiles in self._sols_smi


import hotpot as hp

# initialization Chemical Library
library = Library()

# the public variable
__all__ = ['library']



"""
python v3.7.9
@Project: hotpot
@File   : materials.py
@Author : Zhiyuan Zhang
@Date   : 2023/4/26
@Time   : 1:50
Notes:
    This package is to perform some specific tasks base on the LAMMPS
"""
import os
import os.path as osp
import json
from os.path import join as ptj
from typing import *
import math
import random
import numpy as np
import hotpot
import hotpot.cheminfo as ci


dir_force_field = osp.abspath(ptj(hotpot.data_root, 'force_field'))

# Constants
avogadro = 6.02214076e23  # Avogadro numbers
angstrom = 1e-8  # cm


class AmorphousMaker:
    """ To make Amorphous Materials """
    def __init__(
            self,
            element_composition: Dict[str, float],
            force_field: Union[str, os.PathLike],
            density: float = 1.0,  # g/cm^3
            a: float = 25., b: float = 25., c: float = 25.,
            alpha: float = 90., beta: float = 90., gamma: float = 90.,
    ):
        """"""
        # Check the arguments
        # Determine the path of force field.
        if isinstance(force_field, os.PathLike):
            pff = str(force_field)
        elif osp.exists(force_field):
            pff = force_field
        else:
            pff = ptj(dir_force_field, force_field)

        # Make sure the existence of force field file
        if not osp.exists(pff):
            raise FileNotFoundError('the given force field file is not found!')

        pff = osp.abspath(pff)

        # assign attrs
        sum_freq = sum(f for f in element_composition.values())
        self.elements = {e: f/sum_freq for e, f in element_composition.items()}  # Normalize the elements' frequency
        self.path_force_field = pff
        self.density = density
        self.cryst_params = (a, b, c, alpha, beta, gamma)

    @staticmethod
    def calc_cryst_density(cryst):
        """ Calculation the density for Crystal object """
        mol: ci.Molecule = cryst.molecule
        sum_mass = sum(ci.periodic_table[a.atom_type]['atomic_mass'] for a in mol.atoms)

        # Density, g/cm^3
        return (sum_mass * avogadro) / (cryst.volume * angstrom ** 3)

    @staticmethod
    def density2atom_numbers(ratio_elements: dict, density: float, cryst):
        """
        Calculate the round atom numbers in a crystal.
        Args:
            ratio_elements: Ratio of elements in the crystal
            density: the demand density in the crystal
            cryst: the crystal

        Returns:
            int, the number of atoms
        """
        # Convert the dict of elements and possibility to numpy array
        elements = np.array(list(ratio_elements.keys()))
        possibility = np.array(list(ratio_elements.values()))
        possibility = possibility / possibility.sum()  # Normalize

        average_mol_mass = sum(ci.periodic_table[e]['atomic_mass'] * p for e, p in zip(elements, possibility))

        # TODO: Check ..., the formula may be wrong
        # Terms:   [Mole in Crystal(Total Mass in Crystal (gram)/Average Mole Mass)]/Avogadro Number
        # Units:      g/cm^3     angstrom^3  angstrom/cm             g/mol             _
        num_atom = ((density * (cryst.volume * angstrom ** 3)) / average_mol_mass) * avogadro

        v1, v2, v3 = cryst.vector  # crystal vectors
        dv1, dv2, dv3 = np.sqrt(sum(v1 ** 2)), np.sqrt(sum(v2 ** 2)), np.sqrt(sum(v3 ** 2))  # length of crystal vectors
        min_dv = min(dv1, dv2, dv3)  # the min length in these vectors
        r_dv1, r_dv2, r_dv3 = dv1 / min_dv, dv2 / min_dv, dv3 / min_dv  # the ratios of length of vector to min vector

        # If all vectors are replaced with the smallest one,
        # how many times is the actual crystal volume after replacement
        fold_min_dv_volume = r_dv1 * r_dv2 * r_dv3

        # number of grid points in the direction of min vector
        min_point = math.pow(num_atom / fold_min_dv_volume, 1 / 3)

        # number of point for in the direction of v1, v2, v3
        num_pv1, num_pv2, num_pv3 = int(r_dv1 * min_point), int(r_dv2 * min_point), int(r_dv3 * min_point)
        # the coordinate of grad point in the bases of v1, v2, v3
        pv1, pv2, pv3 = np.meshgrid(
            np.linspace(0, 1, num_pv1, endpoint=False),
            np.linspace(0, 1, num_pv2, endpoint=False),
            np.linspace(0, 1, num_pv3, endpoint=False)
        )

        fraction_coordinates = np.stack((pv1.flatten(), pv2.flatten(), pv3.flatten()))

        # the actual coordinates in the cartesian coordinates
        cartesian_coordinates = np.matmul(cryst.vector, fraction_coordinates).T

        num_atom = len(cartesian_coordinates)
        atomic_symbols = np.random.choice(elements, num_atom, p=possibility)
        atomic_numbers = np.array([ci.periodic_table[symbol]["number"] for symbol in atomic_symbols])

        return atomic_numbers, cartesian_coordinates

    def load_atoms(self):
        mol = ci.Molecule()
        mol.make_crystal(*self.cryst_params)

        cryst = mol.crystal()
        atomic_number, coordinates = self.density2atom_numbers(
            self.elements, self.density, cryst
        )

        mol.quick_build_atoms(atomic_number)
        mol.set(all_coordinates=coordinates.reshape((-1, len(coordinates), 3)))

        mol.configure_select(0)

        return mol

    def melt_quench(
            self, *ff_args, mol=None, path_writefile: Optional[str] = None,
            origin_temp: float = 298.15, melt_temp: float = 4000., highest_temp: float = 10000,
            time_step: float = 0.0001, path_dump_to: Optional[str] = None, dump_every: int = 100,
    ):
        """
        Perform melt-quench process to manufacture a amorphous materials
        Args:
            *ff_args: the arguments the force file requried, refering the LAMMPS pair_coeff:
             "pair_coeff I J args" url: https://docs.lammps.org/pair_coeff.html
            mol: the molecule to be performed melt-quench. if not given, initialize according to elemental
             compositions
            path_writefile: the path to write the final material (screenshot) to file, if not specify, not save.
            origin_temp: the initial temperature before melt
            melt_temp: the round melting point to the materials
            highest_temp: the highest temperature to liquefy the materials
            time_step: time interval between path integrals when performing melt-quench
            path_dump_to: the path to save the trajectory of the melt-quench process, if not specify not save
            dump_every: the step interval between each dumps
        Returns:
            Molecule obj after melt-quench.
        """
        if not isinstance(mol, ci.Molecule):
            mol = self.load_atoms()

        mol.lmp_setup(units='metal')

        # initialization
        mol.lmp.commands_string(
            """
            units metal
            dimension 3
            atom_style full
            """
        )

        # Read molecule into LAMMPS
        mol.lmp.read_main_data()

        # Configure the force field
        mol.lmp("pair_style tersoff")
        mol.lmp(f"pair_coeff * * {self.path_force_field} {' '.join(ff_args)}")

        # Specify the thermodynamical output to screen
        mol.lmp('thermo_style    custom step temp pe etotal press vol density')
        mol.lmp('thermo          1000')

        # the step interval of integral
        mol.lmp(f'timestep {time_step}')

        # Specify the dump configuration
        if path_dump_to:
            dump_fmt = path_dump_to.split('.')[-1]  # the dump fmt is the suffix of file name
            mol.lmp(f'dump mq all {dump_fmt} {dump_every} {path_dump_to}')
            mol.lmp(f'dump_modify mq element {" ".join(set(mol.atomic_symbols))}')

        # Initialize the temperature for system
        mol.lmp(f'velocity all create {origin_temp} {random.randint(100000, 999999)}')

        # Melt
        mol.lmp(f'fix 0 all nvt temp {origin_temp} {highest_temp} 0.7')
        mol.lmp(f'run 10000')

        mol.lmp(f'fix 0 all nvt temp {highest_temp} {highest_temp} 1000')
        while mol.lmp.eval('temp') < highest_temp * 0.95:
            mol.lmp(f'run 1000')

        mol.lmp(f'run 10000')

        # Relax
        mol.lmp('thermo          250')
        mol.lmp(f'fix 0 all nvt temp {melt_temp} {melt_temp} 1000.0')
        while mol.lmp.eval('temp') > melt_temp * 1.05:
            mol.lmp(f'velocity all scale {melt_temp}')
            mol.lmp(f'run 2000')

        mol.lmp('thermo          1000')
        mol.lmp(f'run 20000')

        # Quench
        mol.lmp('thermo          250')
        mol.lmp(f'fix 0 all nvt temp {origin_temp} {origin_temp} 1000.0')
        while mol.lmp.eval('temp') > origin_temp*1.05:
            mol.lmp(f'velocity all scale {(mol.lmp.eval("temp") - origin_temp) / 2 + origin_temp}')
            mol.lmp(f'run 2000')

        if not path_writefile:
            pwf = ptj(os.getcwd(), 'write_dump.xyz')
            write_fmt = 'xyz'
        else:
            pwf = path_writefile
            write_fmt = path_writefile.split('.')[-1]

        mol.lmp(f'write_dump all {write_fmt} {pwf} modify element {" ".join(set(mol.atomic_symbols))}')
        made_mol = ci.Molecule.read_from(pwf)
        if not path_writefile:
            os.remove(pwf)

        made_mol.create_crystal_by_matrix(mol.lmp.cryst_matrix)

        return made_mol




"""
python v3.7.9
@Project: hotpot
@File   : __init__.py.py
@Author : Zhiyuan Zhang
@Date   : 2023/4/26
@Time   : 1:47
"""
import os
import sys

sys.path.append(os.path.abspath(os.path.dirname(__file__)))

from base import HpLammps
from materials import AmorphousMaker



"""
python v3.7.9
@Project: hotpot
@File   : tmo.py
@Author : Zhiyuan Zhang
@Date   : 2023/5/9
@Time   : 5:26
"""
from typing import Sequence
import thermo as tmo
import thermo.chemical as cmc


class Thermo:
    """ To determine thermodynamical properties of Csubstance """
    def __init__(self, mol, **kwargs):
        """"""
        self._data = {
            'mol': mol
        }

        self._data.update(kwargs)
        self._init_property()  # initialization

    def __dir__(self) -> Sequence[str]:
        d = set(self.__dict__.keys())
        chem = self.chem
        eos = self.eos

        if chem:
            d.update(dir(chem))
        if eos:
            d.update(dir(eos))

        return list(d)

    def __getattr__(self, item):
        chem = self._data.get('chem')  # Get chemical substance
        eos = self._data.get('eos_')  # Get equation of state

        if chem:
            attr = getattr(chem, item, None)
            if attr:
                return attr

        if eos:
            attr = getattr(eos, item, None)
            if attr:
                return attr

        return None

    def _init_property(self):
        # Retrieve the substance data by molecule smiles
        T = self._data.get('T')
        P = self._data.get('P')
        V = self._data.get('V')
        chem_kwargs = {}
        if T:
            chem_kwargs['T'] = T
        if P:
            chem_kwargs['P'] = P

        try:
            chem = cmc.Chemical(self.mol.inchi, **chem_kwargs)
            Tc = chem.Tc
            Pc = chem.Pc
            omega = chem.omega

            self._data['chem'] = chem

        except ValueError:
            Tc = self._data.get('Tc')
            Pc = self._data.get('Pc')
            omega = self._data.get('omega')

        # Retrieve the state of equation for thermo.eos module
        eos_name = self._data.get('eos', 'PR')
        eos = getattr(tmo.eos, 'PR')

        if eos and Tc and Pc and omega:
            # if the eos and the critical params are all be retrieved, build eos
            if T and P:
                self._data['eos_'] = eos(T=T, P=P, Tc=Tc, Pc=Pc, omega=omega)
            elif T and V:
                self._data['eos_'] = eos(T=T, V=V, Tc=Tc, Pc=Pc, omega=omega)
            elif P and T:
                self._data['eos_'] = eos(P=P, V=V, Tc=Tc, Pc=Pc, omega=omega)
            else:
                self._data['eos_'] = eos(T=298.15, P=101325, Tc=Tc, Pc=Pc, omega=omega)

    @property
    def chem(self):
        return self._data.get('chem')

    @property
    def eos(self):
        return self._data.get('eos_')

    @property
    def mol(self):
        return self._data.get('mol')

    @property
    def Tc(self):
        return self.chem.Tc

    @property
    def mu_l(self):
        """ TODO: chemical potential in certain state """
        return
